@@ -1,15 +1,23 @@
 use std::cell::RefCell;
+use std::cmp::Ordering;
 use std::collections::{BTreeSet, HashMap, HashSet, VecDeque};
 use std::io::{self, Write};
+use std::rc::Rc;
 
 use anyhow::anyhow;
+use cargo::core::compiler::{CompileKind, RustcTargetData};
+use cargo::core::resolver::features::{ForceAllTargets, HasDevUnits};
+use cargo::core::resolver::CliFeatures;
+use cargo::core::FeatureValue;
 use cargo::core::{dependency::DepKind, Dependency, Package, PackageId, Workspace};
 use cargo::ops::{self, Packages};
+use cargo::util::interning::InternedString;
 use cargo::util::{CargoResult, Config};
 use serde::{Deserialize, Serialize};
-use serde_json;
 use tabwriter::TabWriter;
 
+use crate::error::OutdatedError;
+
 use super::pkg_status::*;
 use super::Options;
 
@@ -29,10 +37,10 @@ pub struct ElaborateWorkspace<'ela> {
 #[derive(Serialize, Deserialize)]
 pub struct CrateMetadata {
     pub crate_name: String,
-    pub dependencies: HashSet<Metadata>,
+    pub dependencies: BTreeSet<Metadata>,
 }
 
-#[derive(Serialize, Deserialize, Eq, Hash, PartialEq)]
+#[derive(Serialize, Deserialize, Eq, PartialEq)]
 pub struct Metadata {
     pub name: String,
     pub project: String,
@@ -42,23 +50,51 @@ pub struct Metadata {
     pub platform: Option<String>,
 }
 
+impl Ord for Metadata {
+    fn cmp(&self, other: &Self) -> Ordering { self.name.cmp(&other.name) }
+}
+
+impl PartialOrd for Metadata {
+    fn partial_cmp(&self, other: &Self) -> Option<Ordering> { Some(self.cmp(other)) }
+}
+
 impl<'ela> ElaborateWorkspace<'ela> {
     /// Elaborate a `Workspace`
     pub fn from_workspace(
         workspace: &'ela Workspace<'_>,
         options: &Options,
     ) -> CargoResult<ElaborateWorkspace<'ela>> {
-        use cargo::core::resolver::ResolveOpts;
+        // new in cargo 0.54.0
+        let flag_features: BTreeSet<FeatureValue> = options
+            .features
+            .iter()
+            .map(|feature| FeatureValue::new(InternedString::from(feature)))
+            .collect();
         let specs = Packages::All.to_package_id_specs(workspace)?;
-        let opts = ResolveOpts::new(
-            true,
-            &options.flag_features,
-            options.all_features(),
-            options.no_default_features(),
-        );
-        let ws_resolve = ops::resolve_ws_with_opts(workspace, opts, &specs)?;
+
+        let cli_features = CliFeatures {
+            features: Rc::new(flag_features),
+            all_features: options.all_features(),
+            uses_default_features: options.no_default_features(),
+        };
+
+        //The CompileKind, this has no target since it's the temp workspace
+        //targets are blank since we don't need to fully build for the targets to get the dependencies
+        let compile_kind = CompileKind::from_requested_targets(workspace.config(), &[])?;
+        let target_data = RustcTargetData::new(workspace, &compile_kind)?;
+        let ws_resolve = ops::resolve_ws_with_opts(
+            workspace,
+            &target_data,
+            &compile_kind,
+            &cli_features,
+            &specs,
+            HasDevUnits::Yes,
+            ForceAllTargets::Yes,
+        )?;
         let packages = ws_resolve.pkg_set;
-        let resolve = ws_resolve.workspace_resolve.expect("Error getting workspace resolved");
+        let resolve = ws_resolve
+            .workspace_resolve
+            .expect("Error getting workspace resolved");
         let mut pkgs = HashMap::new();
         let mut pkg_deps = HashMap::new();
         for pkg in packages.get_many(packages.package_ids())? {
@@ -82,13 +118,13 @@ impl<'ela> ElaborateWorkspace<'ela> {
             pkgs,
             pkg_deps,
             pkg_status: RefCell::new(HashMap::new()),
-            workspace_mode: options.flag_workspace || workspace.current().is_err(),
+            workspace_mode: options.workspace || workspace.current().is_err(),
         })
     }
 
     /// Determine root package based on current workspace and CLI options
     pub fn determine_root(&self, options: &Options) -> CargoResult<PackageId> {
-        if let Some(ref root_name) = options.flag_root {
+        if let Some(ref root_name) = options.root {
             if let Ok(workspace_root) = self.workspace.current() {
                 if root_name == workspace_root.name().as_str() {
                     Ok(workspace_root.package_id())
@@ -98,9 +134,9 @@ impl<'ela> ElaborateWorkspace<'ela> {
                             return Ok(*direct_dep);
                         }
                     }
-                    return Err(anyhow!(
+                    Err(anyhow!(
                         "Root is neither the workspace root nor a direct dependency",
-                    ));
+                    ))
                 }
             } else {
                 Err(anyhow!(
@@ -142,16 +178,19 @@ impl<'ela> ElaborateWorkspace<'ela> {
         dependent_package_name: &str,
     ) -> CargoResult<PackageId> {
         let dependent_package = self.find_contained_package(dependent_package_name)?;
+
         for direct_dep in self.pkg_deps[&dependent_package].keys() {
             if direct_dep.name().as_str() == dependency_name {
                 return Ok(*direct_dep);
             }
         }
+
         for (pkg_id, pkg) in &self.pkgs {
             if pkg.name().as_str() == dependency_name {
                 return Ok(*pkg_id);
             }
         }
+
         Err(anyhow!(
             "Direct dependency {} not found for package {}",
             dependency_name,
@@ -167,6 +206,7 @@ impl<'ela> ElaborateWorkspace<'ela> {
         options: &Options,
         _config: &Config,
         root: PackageId,
+        skip: &HashSet<String>,
     ) -> CargoResult<()> {
         self.pkg_status.borrow_mut().clear();
         let (compat_root, latest_root) = if self.workspace_mode {
@@ -181,7 +221,7 @@ impl<'ela> ElaborateWorkspace<'ela> {
         let mut queue = VecDeque::new();
         queue.push_back((vec![root], Some(compat_root), Some(latest_root)));
         while let Some((path, compat_pkg, latest_pkg)) = queue.pop_front() {
-            let pkg = path.last().unwrap();
+            let pkg = path.last().ok_or(OutdatedError::EmptyPath)?;
             let depth = path.len() as i32 - 1;
             // generate pkg_status
             let status = PkgStatus {
@@ -199,10 +239,12 @@ impl<'ela> ElaborateWorkspace<'ela> {
             );
             self.pkg_status.borrow_mut().insert(path.clone(), status);
             // next layer
-            if options.flag_depth.is_none() || depth < options.flag_depth.unwrap() {
+            // this unwrap is safe since we first check if it is None :)
+            if options.depth.is_none() || depth < options.depth.unwrap() {
                 self.pkg_deps[pkg]
                     .keys()
                     .filter(|dep| !path.contains(dep))
+                    .filter(|&dep| !skip.contains(dep.name().as_str()))
                     .for_each(|&dep| {
                         let name = dep.name();
                         let compat_pkg = compat_pkg
@@ -231,18 +273,24 @@ impl<'ela> ElaborateWorkspace<'ela> {
         options: &Options,
         root: PackageId,
         preceding_line: bool,
+        skip: &HashSet<String>,
     ) -> CargoResult<i32> {
         let mut lines = BTreeSet::new();
         let mut queue = VecDeque::new();
         queue.push_back(vec![root]);
         while let Some(path) = queue.pop_front() {
-            let pkg = path.last().unwrap();
+            let pkg = path.last().ok_or(OutdatedError::EmptyPath)?;
+            let name = pkg.name().to_string();
+
+            if options.ignore.contains(&name) {
+                continue;
+            }
+
             let depth = path.len() as i32 - 1;
             // generate lines
             let status = &self.pkg_status.borrow_mut()[&path];
             if (status.compat.is_changed() || status.latest.is_changed())
-                && (options.flag_packages.is_empty()
-                    || options.flag_packages.contains(&pkg.name().to_string()))
+                && (options.packages.is_empty() || options.packages.contains(&name))
             {
                 // name version compatible latest kind platform
                 let parent = path.get(path.len() - 2);
@@ -251,9 +299,9 @@ impl<'ela> ElaborateWorkspace<'ela> {
                     let label = if self.workspace_mode
                         || parent == &self.workspace.current()?.package_id()
                     {
-                        pkg.name().to_string()
+                        name
                     } else {
-                        format!("{}->{}", self.pkgs[parent].name(), pkg.name())
+                        format!("{}->{}", self.pkgs[parent].name(), name)
                     };
                     let line = format!(
                         "{}\t{}\t{}\t{}\t{:?}\t{}\n",
@@ -271,7 +319,7 @@ impl<'ela> ElaborateWorkspace<'ela> {
                 } else {
                     let line = format!(
                         "{}\t{}\t{}\t{}\t---\t---\n",
-                        pkg.name(),
+                        name,
                         pkg.version(),
                         status.compat.to_string(),
                         status.latest.to_string()
@@ -280,7 +328,8 @@ impl<'ela> ElaborateWorkspace<'ela> {
                 }
             }
             // next layer
-            if options.flag_depth.is_none() || depth < options.flag_depth.unwrap() {
+            // this unwrap is safe since we first check if it is None :)
+            if options.depth.is_none() || depth < options.depth.unwrap() {
                 self.pkg_deps[pkg]
                     .keys()
                     .filter(|dep| !path.contains(dep))
@@ -288,6 +337,7 @@ impl<'ela> ElaborateWorkspace<'ela> {
                         !self.workspace_mode
                             || !self.workspace.members().any(|mem| &mem.package_id() == dep)
                     })
+                    .filter(|&dep| !skip.contains(dep.name().as_str()))
                     .for_each(|&dep| {
                         let mut path = path.clone();
                         path.push(dep);
@@ -314,47 +364,56 @@ impl<'ela> ElaborateWorkspace<'ela> {
                 write!(&mut tw, "{}", line)?;
             }
             tw.flush()?;
-            write!(
-                io::stdout(),
-                "{}",
-                String::from_utf8(tw.into_inner().unwrap()).unwrap()
-            )?;
+            write!(io::stdout(), "{}", String::from_utf8(tw.into_inner()?)?)?;
             io::stdout().flush()?;
         }
 
         Ok(lines.len() as i32)
     }
 
-    pub fn print_json(&'ela self, options: &Options, root: PackageId) -> CargoResult<i32> {
+    pub fn print_json(
+        &'ela self,
+        options: &Options,
+        root: PackageId,
+        skip: &HashSet<String>,
+    ) -> CargoResult<i32> {
         let mut crate_graph = CrateMetadata {
             crate_name: root.name().to_string(),
-            dependencies: HashSet::new(),
+            dependencies: BTreeSet::new(),
         };
         let mut queue = VecDeque::new();
         queue.push_back(vec![root]);
 
         while let Some(path) = queue.pop_front() {
-            let pkg = path.last().unwrap();
+            let pkg = path.last().ok_or(OutdatedError::EmptyPath)?;
+            let name = pkg.name().to_string();
+
+            if options.ignore.contains(&name) {
+                continue;
+            }
+
             let depth = path.len() as i32 - 1;
             // generate lines
             let status = &self.pkg_status.borrow_mut()[&path];
             if (status.compat.is_changed() || status.latest.is_changed())
-                && (options.flag_packages.is_empty()
-                    || options.flag_packages.contains(&pkg.name().to_string()))
+                && (options.packages.is_empty() || options.packages.contains(&name))
             {
                 // name version compatible latest kind platform
-                let parent = path.get(path.len() - 2);
+                // safely get the parent index
+                let parent = if path.len() > 1 {
+                    path.get(path.len() - 2)
+                } else {
+                    None
+                };
 
-                let line;
-
-                if let Some(parent) = parent {
+                let line = if let Some(parent) = parent {
                     let dependency = &self.pkg_deps[parent][pkg];
                     let label = if self.workspace_mode
                         || parent == &self.workspace.current()?.package_id()
                     {
-                        pkg.name().to_string()
+                        name
                     } else {
-                        format!("{}->{}", self.pkgs[parent].name(), pkg.name())
+                        format!("{}->{}", self.pkgs[parent].name(), name)
                     };
 
                     let dependency_type = match dependency.kind() {
@@ -363,29 +422,30 @@ impl<'ela> ElaborateWorkspace<'ela> {
                         DepKind::Build => "Build",
                     };
 
-                    line = Metadata {
+                    Metadata {
                         name: label,
                         project: pkg.version().to_string(),
                         compat: status.compat.to_string(),
                         latest: status.latest.to_string(),
                         kind: Some(dependency_type.to_string()),
                         platform: dependency.platform().map(|p| p.to_string()),
-                    };
+                    }
                 } else {
-                    line = Metadata {
-                        name: pkg.name().to_string(),
+                    Metadata {
+                        name,
                         project: pkg.version().to_string(),
                         compat: status.compat.to_string(),
                         latest: status.latest.to_string(),
                         kind: None,
                         platform: None,
-                    };
-                }
+                    }
+                };
 
                 crate_graph.dependencies.insert(line);
             }
             // next layer
-            if options.flag_depth.is_none() || depth < options.flag_depth.unwrap() {
+            // this unwrap is safe since we first check if it is None :)
+            if options.depth.is_none() || depth < options.depth.unwrap() {
                 self.pkg_deps[pkg]
                     .keys()
                     .filter(|dep| !path.contains(dep))
@@ -396,6 +456,7 @@ impl<'ela> ElaborateWorkspace<'ela> {
                                 .members()
                                 .any(|mem| mem.package_id() == **dep)
                     })
+                    .filter(|&dep| !skip.contains(dep.name().as_str()))
                     .for_each(|dep| {
                         let mut path = path.clone();
                         path.push(*dep);
@@ -404,7 +465,7 @@ impl<'ela> ElaborateWorkspace<'ela> {
             }
         }
 
-        serde_json::to_writer(io::stdout(), &crate_graph)?;
+        println!("{}", serde_json::to_string(&crate_graph)?);
 
         Ok(crate_graph.dependencies.len() as i32)
     }
