@@ -6,17 +6,18 @@ use std::io::{Read, Write};
 use std::path::{Path, PathBuf};
 use std::rc::Rc;
 
-use anyhow::anyhow;
+use anyhow::{anyhow, Context};
 use cargo::core::{Dependency, PackageId, Summary, Verbosity, Workspace};
 use cargo::ops::{update_lockfile, UpdateOptions};
-use cargo::util::errors::CargoResultExt;
+use cargo::sources::config::SourceConfigMap;
 use cargo::util::{CargoResult, Config};
-use semver::{Identifier, Version, VersionReq};
+use semver::{Version, VersionReq};
 use tempfile::{Builder, TempDir};
 use toml::value::Table;
 use toml::Value;
 
 use super::{ElaborateWorkspace, Manifest};
+use crate::error::OutdatedError;
 use crate::Options;
 
 /// A temporary project
@@ -27,6 +28,7 @@ pub struct TempProject<'tmp> {
     config: Config,
     relative_manifest: String,
     options: &'tmp Options,
+    is_workspace_project: bool,
 }
 
 impl<'tmp> TempProject<'tmp> {
@@ -39,7 +41,6 @@ impl<'tmp> TempProject<'tmp> {
         // e.g. /path/to/project
         let workspace_root = orig_workspace.workspace.root();
         let workspace_root_str = workspace_root.to_string_lossy();
-
         let temp_dir = Builder::new().prefix("cargo-outdated").tempdir()?;
         let manifest_paths = manifest_paths(orig_workspace)?;
         let mut tmp_manifest_paths = vec![];
@@ -49,7 +50,7 @@ impl<'tmp> TempProject<'tmp> {
             let mut from_dir = from.clone();
             from_dir.pop();
             let from_dir_str = from_dir.to_string_lossy();
-            
+
             // e.g. /tmp/cargo.xxx/src/sub
             let mut dest = if workspace_root_str.len() < from_dir_str.len() {
                 temp_dir
@@ -58,15 +59,15 @@ impl<'tmp> TempProject<'tmp> {
             } else {
                 temp_dir.path().to_owned()
             };
-            
+
             fs::create_dir_all(&dest)?;
-            
+
             // e.g. /tmp/cargo.xxx/src/sub/Cargo.toml
             dest.push("Cargo.toml");
             tmp_manifest_paths.push(dest.clone());
             fs::copy(from, &dest)?;
 
-            //removing default-run key if it exists to check dependencies 
+            //removing default-run key if it exists to check dependencies
             let mut om: Manifest = {
                 let mut buf = String::new();
                 let mut file = File::open(&dest)?;
@@ -77,7 +78,11 @@ impl<'tmp> TempProject<'tmp> {
             if om.package.contains_key("default-run") {
                 om.package.remove("default-run");
                 let om_serialized = ::toml::to_string(&om).expect("Cannot format as toml file");
-                let mut cargo_toml = OpenOptions::new().read(true).write(true).truncate(true).open(&dest)?;
+                let mut cargo_toml = OpenOptions::new()
+                    .read(true)
+                    .write(true)
+                    .truncate(true)
+                    .open(&dest)?;
                 write!(cargo_toml, "{}", om_serialized)?;
             }
 
@@ -86,14 +91,22 @@ impl<'tmp> TempProject<'tmp> {
             if om.package.contains_key("links") {
                 om.package.remove("links");
                 let om_serialized = ::toml::to_string(&om).expect("Cannot format as toml file");
-                let mut cargo_toml = OpenOptions::new().read(true).write(true).truncate(true).open(&dest)?;
+                let mut cargo_toml = OpenOptions::new()
+                    .read(true)
+                    .write(true)
+                    .truncate(true)
+                    .open(&dest)?;
                 write!(cargo_toml, "{}", om_serialized)?;
             }
 
             if om.package.contains_key("build") {
                 om.package.remove("build");
                 let om_serialized = ::toml::to_string(&om).expect("Cannot format as toml file");
-                let mut cargo_toml = OpenOptions::new().read(true).write(true).truncate(true).open(&dest)?;
+                let mut cargo_toml = OpenOptions::new()
+                    .read(true)
+                    .write(true)
+                    .truncate(true)
+                    .open(&dest)?;
                 write!(cargo_toml, "{}", om_serialized)?;
             }
 
@@ -103,8 +116,6 @@ impl<'tmp> TempProject<'tmp> {
                 dest.push("Cargo.lock");
                 fs::copy(lockfile, dest)?;
             }
-
-
         }
 
         // virtual root
@@ -118,15 +129,30 @@ impl<'tmp> TempProject<'tmp> {
             }
         }
 
-        //.cargo/config 
+        //.cargo/config.toml
+        // this is the preferred way
+        // https://doc.rust-lang.org/cargo/reference/config.html
+        if workspace_root.join(".cargo/config.toml").is_file() {
+            fs::create_dir_all(temp_dir.path().join(".cargo"))?;
+            fs::copy(
+                &workspace_root.join(".cargo/config.toml"),
+                temp_dir.path().join(".cargo/config.toml"),
+            )?;
+        }
+
+        //.cargo/config
+        // this is legacy support for config files without the `.toml` extension
         if workspace_root.join(".cargo/config").is_file() {
-            fs::create_dir_all( temp_dir.path().join(".cargo"))?;
-            fs::copy(&workspace_root.join(".cargo/config"), temp_dir.path().join(".cargo/config"))?;
+            fs::create_dir_all(temp_dir.path().join(".cargo"))?;
+            fs::copy(
+                &workspace_root.join(".cargo/config"),
+                temp_dir.path().join(".cargo/config"),
+            )?;
         }
 
         let relative_manifest = String::from(&orig_manifest[workspace_root_str.len() + 1..]);
         let config = Self::generate_config(temp_dir.path(), &relative_manifest, options)?;
-        
+
         Ok(TempProject {
             workspace: Rc::new(RefCell::new(None)),
             temp_dir,
@@ -134,6 +160,7 @@ impl<'tmp> TempProject<'tmp> {
             config,
             relative_manifest,
             options,
+            is_workspace_project: orig_workspace.workspace_mode,
         })
     }
 
@@ -144,7 +171,7 @@ impl<'tmp> TempProject<'tmp> {
     ) -> CargoResult<Config> {
         let shell = ::cargo::core::Shell::new();
         let cwd = env::current_dir()
-            .chain_err(|| "Cargo couldn't get the current directory of the process")?;
+            .with_context(|| "Cargo couldn't get the current directory of the process")?;
 
         let homedir = ::cargo::util::homedir(&cwd).ok_or_else(|| {
             anyhow!(
@@ -154,26 +181,19 @@ impl<'tmp> TempProject<'tmp> {
         })?;
         let mut cwd = Path::new(root).join(relative_manifest);
         cwd.pop();
-        
-        // Check if $CARGO_HOME is set before capturing the config environment 
-        // if it is, set it in the configure options 
-        let cargo_home_path = match std::env::var_os("CARGO_HOME") {
-            Some(path) => Some(std::path::PathBuf::from(path)),
-            None => None
-        };
+
+        // Check if $CARGO_HOME is set before capturing the config environment
+        // if it is, set it in the configure options
+        let cargo_home_path = std::env::var_os("CARGO_HOME").map(std::path::PathBuf::from);
 
         let mut config = Config::new(shell, cwd, homedir);
         config.configure(
             0,
-            if options.flag_verbose > 0 {
-                None
-            } else {
-                Some(true)
-            },
-            options.flag_color.as_deref(),
+            options.verbose == 0,
+            Some(&options.color.to_string().to_ascii_lowercase()),
             options.frozen(),
             options.locked(),
-            false,
+            options.offline,
             &cargo_home_path,
             &[],
             &[],
@@ -189,8 +209,15 @@ impl<'tmp> TempProject<'tmp> {
             to_update: Vec::new(),
             config: &self.config,
             dry_run: false,
+            workspace: self.is_workspace_project,
         };
-        update_lockfile(self.workspace.borrow().as_ref().unwrap(), &update_opts)?;
+        update_lockfile(
+            self.workspace
+                .borrow()
+                .as_ref()
+                .ok_or(OutdatedError::NoWorkspace)?,
+            &update_opts,
+        )?;
         Ok(())
     }
 
@@ -201,9 +228,9 @@ impl<'tmp> TempProject<'tmp> {
         Ok(())
     }
 
-    fn manipulate_dependencies<F>(manifest: &mut Manifest, f: &F) -> CargoResult<()>
+    fn manipulate_dependencies<F>(manifest: &mut Manifest, f: &mut F) -> CargoResult<()>
     where
-        F: Fn(&mut Table) -> CargoResult<()>,
+        F: FnMut(&mut Table) -> CargoResult<()>,
     {
         if let Some(dep) = manifest.dependencies.as_mut() {
             f(dep)?;
@@ -238,6 +265,7 @@ impl<'tmp> TempProject<'tmp> {
         orig_root: P,
         tmp_root: P,
         workspace: &ElaborateWorkspace<'_>,
+        skipped: &mut HashSet<String>,
     ) -> CargoResult<()> {
         let bin = {
             let mut bin = Table::new();
@@ -259,26 +287,27 @@ impl<'tmp> TempProject<'tmp> {
             if let Some(lib) = manifest.lib.as_mut() {
                 lib.insert("path".to_owned(), Value::String("test_lib.rs".to_owned()));
             }
-            Self::manipulate_dependencies(&mut manifest, &|deps| {
+            Self::manipulate_dependencies(&mut manifest, &mut |deps| {
                 Self::replace_path_with_absolute(
+                    self,
                     deps,
                     orig_root.as_ref(),
                     tmp_root.as_ref(),
                     manifest_path,
+                    skipped,
                 )
             })?;
 
             let package_name = manifest.name();
             let features = manifest.features.clone();
-            Self::manipulate_dependencies(&mut manifest, &|deps| {
+            Self::manipulate_dependencies(&mut manifest, &mut |deps| {
                 self.update_version_and_feature(deps, &features, workspace, &package_name, false)
             })?;
 
-
             Self::write_manifest(&manifest, manifest_path)?;
         }
         let root_manifest = self.temp_dir.path().join(&self.relative_manifest);
-        
+
         *self.workspace.borrow_mut() =
             Some(Workspace::new(Path::new(&root_manifest), &self.config)?);
         Ok(())
@@ -290,6 +319,7 @@ impl<'tmp> TempProject<'tmp> {
         orig_root: P,
         tmp_root: P,
         workspace: &ElaborateWorkspace<'_>,
+        skipped: &mut HashSet<String>,
     ) -> CargoResult<()> {
         let bin = {
             let mut bin = Table::new();
@@ -310,19 +340,23 @@ impl<'tmp> TempProject<'tmp> {
             if let Some(lib) = manifest.lib.as_mut() {
                 lib.insert("path".to_owned(), Value::String("test_lib.rs".to_owned()));
             }
-            Self::manipulate_dependencies(&mut manifest, &|deps| {
+            Self::manipulate_dependencies(&mut manifest, &mut |deps| {
                 Self::replace_path_with_absolute(
+                    self,
                     deps,
                     orig_root.as_ref(),
                     tmp_root.as_ref(),
                     manifest_path,
+                    skipped,
                 )
             })?;
+
             let package_name = manifest.name();
             let features = manifest.features.clone();
-            Self::manipulate_dependencies(&mut manifest, &|deps| {
+            Self::manipulate_dependencies(&mut manifest, &mut |deps| {
                 self.update_version_and_feature(deps, &features, workspace, &package_name, true)
             })?;
+
             Self::write_manifest(&manifest, manifest_path)?;
         }
 
@@ -343,14 +377,15 @@ impl<'tmp> TempProject<'tmp> {
         let package_id = workspace.find_direct_dependency(name, dependent_package_name)?;
         let version = package_id.version();
         let source_id = package_id.source_id().with_precise(None);
-        let mut source = source_id.load(&self.config, &HashSet::new())?;
-        if !source_id.is_default_registry() {
-            let _lock = self.config.acquire_package_cache_lock()?;
-            source.update()?;
-        }
-        let dependency = Dependency::parse_no_deprecated(name, None, source_id)?;
         let query_result = {
-            let _lock = self.config.acquire_package_cache_lock()?;
+            let ws_config = workspace.workspace.config();
+            let _lock = ws_config.acquire_package_cache_lock()?;
+            let source_config = SourceConfigMap::new(ws_config)?;
+            let mut source = source_config.load(source_id, &HashSet::new())?;
+            if !source_id.is_default_registry() {
+                source.update()?;
+            }
+            let dependency = Dependency::parse(name, None, source_id)?;
             let mut query_result = source.query_vec(&dependency)?;
             query_result.sort_by(|a, b| b.version().cmp(a.version()));
             query_result
@@ -359,39 +394,55 @@ impl<'tmp> TempProject<'tmp> {
             Some(requirement) => Some(VersionReq::parse(requirement)?),
             None => None,
         };
-        let latest_result = query_result
-            .iter()
-            .find(|summary| {
-                if summary.version() < version {
-                    false
-                } else if version_req.is_none() {
-                    true
-                } else if find_latest {
-                    self.options.flag_aggressive
-                        || valid_latest_version(requirement.unwrap(), summary.version())
-                } else {
-                    version_req.as_ref().unwrap().matches(summary.version())
-                }
-            })
-            .unwrap_or_else(|| {
-                panic!(
-                    "Cannot find matched versions of package {} from source {}",
-                    name, source_id
-                )
-            });
-        Ok(latest_result.clone())
+        let latest_result = query_result.iter().find(|summary| {
+            if summary.version() < version {
+                false
+            } else if version_req.is_none() {
+                true
+            } else if find_latest {
+                // this unwrap is safe since we check if `version_req` is `None` before this
+                // (which is only `None` if `requirement` is `None`)
+                self.options.aggressive
+                    || valid_latest_version(requirement.unwrap(), summary.version())
+            } else {
+                // this unwrap is safe since we check if `version_req` is `None` before this
+                version_req.as_ref().unwrap().matches(summary.version())
+            }
+        });
+
+        let latest_summary = match latest_result {
+            Some(summary) => summary,
+            None => {
+                // If the version_req cannot be found use the version
+                // this happens when we use a git repository as a dependency, without specifying
+                // the version in Cargo.toml, preventing us from needing an unwrap below in the warn
+                let ver_req = match version_req {
+                    Some(v_r) => format!("{}", v_r),
+                    None => format!("{}", version),
+                };
+                // this should be safe it should only fail if we cannot get
+                // access to write to the terminal
+                // if this fails it's a cargo (as a dependency) issue
+                self.warn(format!(
+                    "cannot compare {} crate version found in toml {} with crates.io latest {}",
+                    name,
+                    ver_req,
+                    query_result[0].version()
+                ))?;
+
+                //this returns the latest version
+                &query_result[0]
+            }
+        };
+
+        Ok(latest_summary.clone())
     }
 
     fn feature_includes(&self, name: &str, optional: bool, features_table: &Option<Value>) -> bool {
         if self.options.all_features() {
             return true;
         }
-        if !optional
-            && self
-                .options
-                .flag_features
-                .contains(&String::from("default"))
-        {
+        if !optional && self.options.features.contains(&String::from("default")) {
             return true;
         }
         let features_table = match *features_table {
@@ -400,7 +451,7 @@ impl<'tmp> TempProject<'tmp> {
         };
         let mut to_resolve: Vec<&str> = self
             .options
-            .flag_features
+            .features
             .iter()
             .filter(|f| !f.is_empty())
             .map(String::as_str)
@@ -440,7 +491,19 @@ impl<'tmp> TempProject<'tmp> {
     ) -> CargoResult<()> {
         let dep_keys: Vec<_> = dependencies.keys().cloned().collect();
         for dep_key in dep_keys {
-            let original = dependencies.get(&dep_key).cloned().unwrap();
+            // this, by brute force, allows a user to exclude a dependency by not writing
+            // it to the temp project's manifest
+            // In short this allows cargo to build the package with semver minor compatibilities issues
+            // https://github.com/rust-lang/cargo/issues/6584
+            // https://github.com/kbknapp/cargo-outdated/issues/230
+            if self.options.exclude.contains(&dep_key) {
+                continue;
+            }
+
+            let original = dependencies
+                .get(&dep_key)
+                .cloned()
+                .ok_or(OutdatedError::NoMatchingDependency)?;
 
             match original {
                 Value::String(requirement) => {
@@ -491,7 +554,7 @@ impl<'tmp> TempProject<'tmp> {
                             }
                         })
                         .unwrap_or(false);
-                    if !self.feature_includes(&name, optional, features) {
+                    if !self.feature_includes(name, optional, features) {
                         continue;
                     }
                     let mut replaced = t.clone();
@@ -501,7 +564,11 @@ impl<'tmp> TempProject<'tmp> {
                         _ => None,
                     };
                     let r_summary = self.find_update(
-                        if orig_name == "" { &name } else { &orig_name },
+                        if orig_name.is_empty() {
+                            name
+                        } else {
+                            orig_name
+                        },
                         package_name,
                         requirement,
                         workspace,
@@ -534,6 +601,9 @@ impl<'tmp> TempProject<'tmp> {
                                     };
                                     let retained =
                                         features_and_options(&summary).contains(feature.as_str());
+                                    // this unwrap should be safe it should only fail if we cannot get
+                                    // access to write to the terminal
+                                    // if this fails it's a cargo (as a dependency) issue
                                     if !retained {
                                         self.warn(format!(
                                             "Feature {} of package {} \
@@ -564,14 +634,19 @@ impl<'tmp> TempProject<'tmp> {
     }
 
     fn replace_path_with_absolute(
+        &self,
         dependencies: &mut Table,
         orig_root: &Path,
         tmp_root: &Path,
         tmp_manifest: &Path,
+        skipped: &mut HashSet<String>,
     ) -> CargoResult<()> {
         let dep_names: Vec<_> = dependencies.keys().cloned().collect();
         for name in dep_names {
-            let original = dependencies.get(&name).cloned().unwrap();
+            let original = dependencies
+                .get(&name)
+                .cloned()
+                .ok_or(OutdatedError::NoMatchingDependency)?;
             match original {
                 Value::Table(ref t) if t.contains_key("path") => {
                     if let Value::String(ref orig_path) = t["path"] {
@@ -587,16 +662,30 @@ impl<'tmp> TempProject<'tmp> {
                                 relative.join(orig_path)
                             };
                             if !tmp_root.join(&relative).join("Cargo.toml").exists() {
-                                let mut replaced = t.clone();
-                                replaced.insert(
-                                    "path".to_owned(),
-                                    Value::String(
-                                        fs::canonicalize(orig_root.join(relative))?
-                                            .to_string_lossy()
-                                            .to_string(),
-                                    ),
-                                );
-                                dependencies.insert(name, Value::Table(replaced));
+                                if self.options.root_deps_only {
+                                    dependencies.remove(&name);
+
+                                    if t.contains_key("package") {
+                                        if let Value::String(ref package_name) = t["package"] {
+                                            skipped.insert(package_name.to_string());
+                                        } else {
+                                            skipped.insert(name);
+                                        }
+                                    } else {
+                                        skipped.insert(name);
+                                    }
+                                } else {
+                                    let mut replaced = t.clone();
+                                    replaced.insert(
+                                        "path".to_owned(),
+                                        Value::String(
+                                            fs::canonicalize(orig_root.join(relative))?
+                                                .to_string_lossy()
+                                                .to_string(),
+                                        ),
+                                    );
+                                    dependencies.insert(name, Value::Table(replaced));
+                                }
                             }
                         }
                     }
@@ -609,13 +698,11 @@ impl<'tmp> TempProject<'tmp> {
 
     fn warn<T: ::std::fmt::Display>(&self, message: T) -> CargoResult<()> {
         let original_verbosity = self.config.shell().verbosity();
-        self.config
-            .shell()
-            .set_verbosity(if self.options.flag_quiet {
-                Verbosity::Quiet
-            } else {
-                Verbosity::Normal
-            });
+        self.config.shell().set_verbosity(if self.options.quiet {
+            Verbosity::Quiet
+        } else {
+            Verbosity::Normal
+        });
         self.config.shell().warn(message)?;
         self.config.shell().set_verbosity(original_verbosity);
         Ok(())
@@ -654,23 +741,32 @@ fn manifest_paths(elab: &ElaborateWorkspace<'_>) -> CargoResult<Vec<PathBuf>> {
         visited.insert(pkg_id);
         let pkg = &elab.pkgs[&pkg_id];
         let pkg_path = pkg.root().to_string_lossy();
-        let cargo_home_path = match std::env::var_os("CARGO_HOME") {
-             Some(path) => path.into_string().expect("Error getting string from OsString"),
-             None => "".to_string(),
-         };
 
-         if pkg_path.starts_with(workspace_path) {
-             if !pkg_path.starts_with(&cargo_home_path){
-                 manifest_paths.push(pkg.manifest_path().to_owned());
-             }
-         }        
+        // Checking if there's a CARGO_HOME set and that it is not an empty string
+        let cargo_home_path = match std::env::var_os("CARGO_HOME") {
+            Some(path) if !path.is_empty() => Some(
+                path.into_string()
+                    .expect("Error getting string from OsString"),
+            ),
+            _ => None,
+        };
+
+        // If there is a CARGO_HOME make sure we do not crawl the registry for more Cargo.toml files
+        // Otherwise add all Cargo.toml files to the manifest paths
+        if pkg.root().starts_with(PathBuf::from(workspace_path))
+            && (cargo_home_path.is_none()
+                || !pkg_path
+                    .starts_with(&cargo_home_path.expect("Error extracting CARGO_HOME string")))
+        {
+            manifest_paths.push(pkg.manifest_path().to_owned());
+        }
 
         for &dep in elab.pkg_deps[&pkg_id].keys() {
             manifest_paths_recursive(dep, elab, workspace_path, visited, manifest_paths)?;
         }
 
         Ok(())
-    };
+    }
 
     // executed against a virtual manifest
     let workspace_path = elab.workspace.root().to_string_lossy();
@@ -690,7 +786,7 @@ fn manifest_paths(elab: &ElaborateWorkspace<'_>) -> CargoResult<Vec<PathBuf>> {
 }
 
 fn valid_latest_version(mut requirement: &str, version: &Version) -> bool {
-    match (requirement.contains('-'), version.is_prerelease()) {
+    match (requirement.contains('-'), !version.pre.is_empty()) {
         // if user was on a stable channel, it's unlikely for him to update to an unstable one
         (false, true) => false,
         // both are stable, leave for further filters
@@ -699,14 +795,19 @@ fn valid_latest_version(mut requirement: &str, version: &Version) -> bool {
         (false, false) | (true, false) => true,
         // both are unstable, must be in the same channel
         (true, true) => {
-            requirement = requirement.trim_start_matches(&['=',' ','~','^'][..]);
-            let requirement_version = Version::parse(&requirement).expect("Error could not parse requirement into a semantic version"); 
-            let requirement_channel = requirement_version.pre[0].to_string();
-            match (requirement_channel.is_empty(), &version.pre[0]) {
-                (true, &Identifier::Numeric(_)) => true,
-                (false, &Identifier::AlphaNumeric(_)) => {
-                    Identifier::AlphaNumeric(requirement_channel) == version.pre[0]
-                }
+            requirement = requirement.trim_start_matches(&['=', ' ', '~', '^'][..]);
+            let requirement_version = Version::parse(requirement)
+                .expect("Error could not parse requirement into a semantic version");
+            let requirement_channel = requirement_version.pre.split('.').next().unwrap();
+            let requirement_channel_numeric =
+                requirement_channel.bytes().all(|b| b.is_ascii_digit());
+
+            let version_channel = version.pre.split('.').next().unwrap();
+            let version_channel_numeric = version_channel.bytes().all(|b| b.is_ascii_digit());
+
+            match (requirement_channel_numeric, version_channel_numeric) {
+                (false, false) => requirement_channel == version_channel,
+                (true, true) => true,
                 _ => false,
             }
         }
